(ns
  ^{:author "Brian Craft"
    :doc "xena dataset loader entry point."}
  cavm.loader
  (:require [me.raynes.fs :as fs])
  (:require [cavm.fs-utils :refer [relativize]])
  (:require [cavm.multireader :refer [multi-reader with-multi]])
  (:require [clj-time.coerce :refer [from-long]])
  (:require [clojure.java.io :as io])
  (:require [digest])
  (:require [clojure.tools.logging :refer [info warn]])
  (:require [clojure.data.priority-map :refer [priority-map]])
  (:require [cavm.dir-watch :refer [dir-change-queue dir-watcher]])
  (:require [cavm.db :as cdb]))

; Combines file readers with db writers.

(defn- file-time [file]
  (from-long (.lastModified (java.io.File. file))))

(defn- file-hash [file]
  (digest/sha1 (io/as-file file)))

(defn- file-stats [docroot file]
  (let [full-path (str (io/file docroot file))
        ts (file-time full-path)
        h (file-hash full-path)]
    {:name file :time ts :hash h}))

(defn- log-error [filename e]
  (warn e "Loading" filename))

(defn now []
  (.getTime (java.util.Date.)))

(defn delay-cb [cb to]
  (future 
    (do (Thread/sleep to) (cb))))

(defn queue-with-delay*
  "Queue a function to run delayed by a fixed period after an event. Returns
  a function to be called on the event. The delay is reset if a second event
  occurs."
  [period func & args]
  (let [q (agent nil)]
    (fn []
      (send q (fn [t]
                    (when t
                      (future-cancel t))
                    (future (Thread/sleep period) (send q (fn [_] (apply func args) nil))))))))

(defmacro queue-with-delay
  [period & body]
  `(queue-with-delay* ~period (fn [] ~@body)))

; - get first file on queue
; - if old enough, update loadModified
; - if still old enough, run
; - otherwise put back on queue & reschedule

;(defn update-queue [q t quiet]
;  (let [maybe-ready (take-while (fn [[file ts]] (> quiet (- t ts))) @q)
;        updates (map (fn [[file _]] [file (.lastModified file)]) maybe-ready)]
;    (when (not (= maybe-ready updates))
;      (swap! q into updates))))

(defn run-queue
  "Loader thread loop.

  Load queued file older than quiet period and reschedule loader; or,
  schedule loader after timeout if queue not empty but files too new; or,
  don't schedule loader (queue API will schedule next load)."
  [_ l-agent loader waker q quiet]
  (let [t (now)
        [file ts] (first @q)
        reschedule (fn [] (send-off l-agent run-queue l-agent loader waker q quiet))]
    (when ts
      (if (>= (- t ts) quiet)
        (do
          (swap! q pop)
          (loader file)
          (when (count @q)
            (reschedule)))
        (do 
          (when-let [w @waker]
            (future-cancel w))
          (reset! waker (delay-cb
                          reschedule
                          (- (+ ts quiet) t)))))) ; non-negative due to the conditional
    nil))

; t - ts >= quiet
; 0 >= quiet + ts - t
; if ts is in the future
; ts t  (ts + quiet)
(defn queue-watcher [cb dir & [{:keys [quiet interval] :or {quiet (* 5 1000)
                                                            interval (* 1 1000)}}]]
  (let [l-agent (agent nil)
        queue (atom (priority-map))
        waker (atom nil)
        watcher (dir-watcher
                  (fn [{:keys [new old changed files]}]
                    (swap! queue into (select-keys files (concat new changed)))
                    (println "new queue " @queue)
                    (send-off l-agent run-queue l-agent cb waker queue quiet))
                  dir
                  {:interval interval})]
    (println interval)
    (set-error-handler! l-agent (fn [a e] (warn e "Loader agent")))
    {:stop watcher
     :add (fn [file] (do (swap! queue assoc file (.lastModified file))
                       (send-off l-agent run-queue l-agent cb waker queue quiet)))}))

(defn- write-matrix [db docroot filename reader always]
  (let [{:keys [metadata refs features data-fn]} reader
        dependencies (if features
                       [(refs ":clinicalFeature") filename]
                       [filename])
        files (mapv #(file-stats docroot %) dependencies)]
    (with-multi [in (multi-reader (io/file docroot filename))]
      (cdb/write-matrix
        db
        filename
        files
        metadata
        (partial data-fn in)
        features
        always))))

(defn- delete-matrix [db docroot filename]
  (let [fname (str (relativize docroot filename))]
    (cdb/delete-matrix db fname)))

(def ^:private loaders
  {:probemap write-matrix
   :matrix write-matrix
   :mutation write-matrix
   :segment write-matrix})

(defn- ignore [& args])

; XXX The reader should really return a lazy seq of data types and readers, since
; one file type may contain mutiple xena data types.
(defn loader
  "Load a data file into the database"
  [db detector docroot filename & [opts]]
  (let [fname  (str (relativize docroot filename))
        reader @(:reader (detector filename))
        loader (get loaders (:datatype reader) ignore)]
    (loader db docroot fname reader opts)))

(defn loader-agent
  "Create an agent to serialize bulk loads, returning a function
  for loading a file via the agent.

  fn [filename & [{:keys [always delete] :or {always false delete false}}]]

  Detector is a function returning a file type as a keyword, which
  will be used to select a dataset writer method. This has evolved
  to a noop, as we currently have a single dataset writer."

  [db detector docroot]
  (let [a (agent nil)]
    (fn [filename & [{:keys [delete] :as opts}]]
      (let [filename (fs/with-cwd docroot (fs/file filename))]
        (if delete
          (send-off a (fn [n]
                        (info "Removing dataset" (str filename))
                        (let [t (read-string
                                  (with-out-str
                                    (time (try
                                            (delete-matrix db docroot filename)
                                            (catch Exception e (log-error filename e))))))]
                          (info (str "Removed " filename ", " t)))
                        nil))
          (send-off a (fn [n]
                        (info "Loading dataset" (str filename))
                        (let [t (read-string
                                  (with-out-str
                                    (time (try
                                            (loader db detector docroot filename opts)
                                            (catch Exception e (log-error filename e))))))]
                          (info (str "Loaded " filename ", " t)))
                        nil)))))))
